import json
from datetime import datetime
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models.fields.files import FieldFile


DEFAULT_TIMEZONE = getattr(settings, "TIMEZONE", "UTC")


def assure_tz(dt, tz=DEFAULT_TIMEZONE):
    """Add timezone information to a datetime if not already present"""
    if not dt:
        return dt
    if not dt.tzinfo:
        dt = tz.localize(dt)
    return dt


class JSONEncoder(DjangoJSONEncoder):
    """A JSON encoder with additional serialiation options
    This encoder serializes:
     - `datetime`s as isoformatted, timezone-aware strings
     - django `FieldFile`s by their URL string
     - python `set`s as lists
    """

    def default(self, o):
        """Override default serialization to apply isoformat to timezones"""
        if o is None:
            return None
        if isinstance(o, datetime):
            value = assure_tz(o.astimezone())
            return value.isoformat()
        if issubclass(o.__class__, FieldFile):
            return o.url if bool(o) else None
        if isinstance(o, set):
            return list(o)
        return super().default(o)


def serialize(value):
    return json.dumps(value, cls=JSONEncoder)


def deserialize(value):
    return json.loads(value)



import json
import logging
from django.conf import settings
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from .signals import event_received


logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name="dispatch")
class GoogleCloudEventsView(View):
    """Handles events inbound from Google Cloud services like Pub/Sub by dispatch to a django signal

    Any exceptions thrown by the handlers will be returned to the client as 400s.

    """

    def post(self, request, event_kind, event_reference):
        """Handle a POSTed event"""
        try:
            event_payload = json.loads(request.body)
            event_parameters = request.GET.dict()
            event_received.send(
                sender=self.__class__,
                event_kind=event_kind,
                event_reference=event_reference,
                event_payload=event_payload,
                event_parameters=event_parameters,
            )
            return self._prepare_response(status=201, payload={})

        except Exception as e:  # pylint: disable=broad-except
            if getattr(settings, "DEBUG", False):
                raise e
            else:
                msg = f"Unable to handle event of kind {event_kind} with reference {event_reference}"
                logger.warning("%s. Exception: %s", msg, str(e))
                return self._prepare_response(status=400, payload={"error": msg})

    def _prepare_response(self, status, payload):
        return HttpResponse(status=status, content=json.dumps(payload), content_type="application/json")



from typing import List


class UnsupportedFormatException(Exception):
    pass


class ValidationError(Exception):
    pass


class UnboundException(Exception):
    pass


class NotFound(Exception):
    pass


class FailedPrecondition(Exception):
    pass


class DeletedRecently(FailedPrecondition):
    def __init__(self, resource, blocked_period="1 week"):
        message = f"{resource} was probably deleted recently. Cannot reuse name for {blocked_period}."
        super().__init__(message)


class BigQueryJobError(Exception):
    def __init__(self, job):
        message = " | ".join([error["message"] for error in job.errors])
        super().__init__(message)


class AlreadyExists(Exception):
    pass


class AlreadyDeleted(Exception):
    pass


class NotAllowed(Exception):
    pass


class InvalidPassword(Exception):
    pass


class MissingUserIdentification(Exception):
    pass


class PushWebhookInvalid(ValidationError):
    pass


class ChannelIdNotUnique(Exception):
    pass


class NotACalendarUser(Exception):
    pass


class QuotaExceeded(Exception):
    pass


class OperationError(Exception):
    def __init__(self, errors: List):
        self.errors = errors
        super().__init__(str(self.errors))



# TODO: Refactor tests to import members directly then remove this module export
from . import gcloud
from .gcloud import GoogleCloudFile, GoogleCloudMediaStorage, GoogleCloudStaticStorage, GoogleCloudStorage


__all__ = ["GoogleCloudStorage", "GoogleCloudFile", "gcloud", "GoogleCloudMediaStorage", "GoogleCloudStaticStorage"]



import logging
from django.forms.fields import JSONField

from .widgets import CloudObjectWidget


logger = logging.getLogger(__name__)


class CloudObjectFormField(JSONField):
    """Overrides the forms.JSONField with a widget to handle file selection and upload"""

    empty_value = None
    widget = CloudObjectWidget



from functools import cached_property
from urllib.error import URLError
from urllib.request import Request, urlopen

from ..exceptions import NotOnCloudRunError


ENDPOINTS = {
    "project_id": "computeMetadata/v1/project/project-id",
    "project_number": "/computeMetadata/v1/project/numeric-project-id",
    "region": "/computeMetadata/v1/instance/region",
    "compute_instance_id": "/computeMetadata/v1/instance/id",
    "email": "/computeMetadata/v1/instance/service-accounts/default/email",
    "token": "/computeMetadata/v1/instance/service-accounts/default/token",
}


class CloudRunMetadata:
    """A helper class for fetching Cloud Run internal service metadata

    Based on: https://cloud.google.com/run/docs/container-contract#metadata-server

    Works while running on Cloud Run instances, excepts when not running under
    Google infrastructure.

    The email / token properties refer to / use the *runtime* service account,
    NOT the *default* service account as is suggested by the endpoint naming
    convention. See https://stackoverflow.com/questions/75770844/does-this-cloud-run-metadata-server-endpoint-provide-the-default-service-account?noredirect=1#comment133690781_75770844
    """

    def _fetch(self, endpoint):
        url = f"http://metadata.google.internal/{endpoint}"
        req = Request(url)
        req.add_header("Metadata-Flavor", "Google")
        try:
            return urlopen(req).read().decode()
        except URLError as exc:
            raise NotOnCloudRunError(
                "Attempted to fetch GCP-internal metadata, this can only be done when being run on Google Compute Engine services"
            ) from exc

    @cached_property
    def is_cloud_run(self):
        """True if currently running on Cloud Run infrastructure, False otherwise"""
        try:
            self._fetch(ENDPOINTS["project-id"])
            return True
        except NotOnCloudRunError:
            return False

    @cached_property
    def project_id(self):
        """Project ID of the project the Cloud Run service or job belongs to"""
        return self._fetch(ENDPOINTS["project_id"])

    @cached_property
    def project_number(self):
        """Project number of the project the Cloud Run service or job belongs to"""
        return self._fetch(ENDPOINTS["project_number"])

    @cached_property
    def region(self):
        """Region of this Cloud Run service or job"""
        # GCP Metadata server returns projects/PROJECT-NUMBER/regions/REGION
        return self._fetch(ENDPOINTS["region"]).split("/")[-1]

    @cached_property
    def container_instance_id(self):
        """Unique identifier of the container instance (also available in logs)"""
        return self._fetch(ENDPOINTS["container_instance_id"])

    @cached_property
    def email(self):
        """Email for the runtime service account of this Cloud Run service or job."""
        return self._fetch(ENDPOINTS["email"])

    @property
    def token(self):
        """Generates an OAuth2 access token for the service account of this Cloud Run service or job.
        The Cloud Run service agent is used to fetch a token. This endpoint will
        return a JSON response with an access_token attribute.

        Unlike the other endpoints, `token` is NOT cached since the token may rotate within
        the container lifespan.
        """
        return self._fetch(ENDPOINTS["token"])
